
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

            #include <glib-object.h>
            #include <gst/controller/controller-prelude.h>

            G_BEGIN_DECLS

/* enumerations from "gstinterpolationcontrolsource.h" */

GST_CONTROLLER_API
GType gst_interpolation_mode_get_type (void);
#define GST_TYPE_INTERPOLATION_MODE (gst_interpolation_mode_get_type())

/* enumerations from "gstlfocontrolsource.h" */

GST_CONTROLLER_API
GType gst_lfo_waveform_get_type (void);
#define GST_TYPE_LFO_WAVEFORM (gst_lfo_waveform_get_type())

G_END_DECLS

/* Generated data ends here */

